/*
*
* avl.c -- C source file for avl trees.
*
* Created 03/01/89 by Brad Appleton
*
* Fri Jul 14 13:53:42 1989, Rev 1.0, brad(0165)
*
* Completely reformatted and reworked BLP 7/19/96.
* Bugs introduced 8/12/95 corrected 12/11/95.
* Hacked for PSPP by BLP 7/21/95, 8/12/95.
*
**/

#include <config.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "avl.h"
#include "common.h"
#include "misc.h"

/* Directional Definitions. */
#define OPPOSITE(x)     (1 - (x))
typedef enum
  {
    AVL_LEFT,
    AVL_RIGHT,
  }
DIRECTION;

/* Return codes used by avl_insert(), avl_delete(), and balance(). */
enum
  {
    HEIGHT_UNCHANGED = 0,
    HEIGHT_CHANGED = 1,
  };

/* Balance definitions. */
#define LEFT_IMBALANCE(nd) ((nd)->bal < LEFT_HEAVY)
#define RIGHT_IMBALANCE(nd) ((nd)->bal > RIGHT_HEAVY)

enum
  {
    LEFT_HEAVY = -1,
    BALANCED = 0,
    RIGHT_HEAVY = 1,
  };

/* The tree we're currently dealing with. */
static avl_tree *cur_tree;

/* Auxiliary functions. */

/* Passed to avl_destroy, frees a node. */
void
free_func (void *p, VISIT order, void *param)
{
  free (p);
}

/* Creates a new node with data DATA in arena A or with xmalloc if A
   is NULL. */
static avl_node *
new_node (void *data)
{
  avl_node *root;

  if (cur_tree->owner)
    root = (avl_node *) arena_alloc (cur_tree->owner, sizeof (avl_node));
  else
    root = (avl_node *) xmalloc (sizeof (avl_node));
  root->data = data;
  root->bal = BALANCED;
  root->subtree[AVL_LEFT] = root->subtree[AVL_RIGHT] = NULL;

  return root;
}

/* Returns node type for given node TREE. */
static NODE
node_type (avl_node * tree)
{
  if (tree == NULL)
    return IS_NULL;
  else if (tree->subtree[AVL_LEFT] && tree->subtree[AVL_RIGHT])
    return IS_TREE;
  else if (tree->subtree[AVL_LEFT])
    return IS_LBRANCH;
  else if (tree->subtree[AVL_RIGHT])
    return IS_RBRANCH;
  else
    return IS_LEAF;
}

/* Private functions. */

/* Compare function used to find the minimal element in a tree. */
static int
min_node (void *elt1, void *elt2, void *param, NODE nd_typ)
{
  /* Trace all the way down the left subtrees. */
  return (nd_typ == IS_RBRANCH || nd_typ == IS_LEAF) ? 0 : -1;
}

/* Compare function used to find the maximal element in a tree. */
static int
max_node (void *elt1, void *elt2, void *param, NODE nd_typ)
{
  /* Trace all the way down the right subtrees. */
  return (nd_typ == IS_LBRANCH || nd_typ == IS_LEAF) ? 0 : 1;
}

/* Rotates node ROOTP in direction DIR. */
int
rotate_once (avl_node ** rootp, DIRECTION dir)
{
  DIRECTION other_dir = OPPOSITE (dir);		/* Opposite direction to `dir'. */
  avl_node *old_root = *rootp;	/* Copy of original root of tree. */
  short ht_unchanged;		/* True if height unchanged. */

  ht_unchanged = !((*rootp)->subtree[other_dir]->bal);

  /* Assign new root. */
  *rootp = old_root->subtree[other_dir];

  /* New root exchanges its `dir' subtree for its parent. */
  old_root->subtree[other_dir] = (*rootp)->subtree[dir];
  (*rootp)->subtree[dir] = old_root;

  /* Update balances. */
  old_root->bal = -(dir == AVL_LEFT ? --((*rootp)->bal) : ++((*rootp)->bal));

  return ht_unchanged;
}

/* Rotates node ROOTP in direction DIR then in the opposite
   direction. */
static void
rotate_twice (avl_node ** rootp, DIRECTION dir)
{
  DIRECTION other_dir = OPPOSITE (dir);
  avl_node *old_root = *rootp;
  avl_node *old_other_dir_subtree = (*rootp)->subtree[other_dir];

  /* Assign new root. */
  *rootp = old_root->subtree[other_dir]->subtree[dir];

  /* New root exchanges its `dir' subtree for its grandparent. */
  old_root->subtree[other_dir] = (*rootp)->subtree[dir];
  (*rootp)->subtree[dir] = old_root;

  /* New root exchanges its `other-dir' subtree for its parent. */
  old_other_dir_subtree->subtree[dir] = (*rootp)->subtree[other_dir];
  (*rootp)->subtree[other_dir] = old_other_dir_subtree;

  /* Update balances. */
  (*rootp)->subtree[AVL_LEFT]->bal = -max ((*rootp)->bal, 0);
  (*rootp)->subtree[AVL_RIGHT]->bal = -min ((*rootp)->bal, 0);
  (*rootp)->bal = 0;
}

/* Restores the balance of tree ROOTP.  Returns 1 if tree height
   changed due to rotation, 0 otherwise. */
int
balance (avl_node ** rootp)
{
  int special_case = 0;

  if (LEFT_IMBALANCE (*rootp))
    {
      /* Right rotation. */
      if ((*rootp)->subtree[AVL_LEFT]->bal == RIGHT_HEAVY)
	rotate_twice (rootp, AVL_RIGHT);	/* Double RL. */
      else
	special_case = rotate_once (rootp, AVL_RIGHT);	/* Single RR. */
    }
  else if (RIGHT_IMBALANCE (*rootp))
    {
      /* Left rotation. */
      if ((*rootp)->subtree[AVL_RIGHT]->bal == LEFT_HEAVY)
	rotate_twice (rootp, AVL_LEFT);		/* Double LR. */
      else
	special_case = rotate_once (rootp, AVL_LEFT);	/* Single LL. */
    }
  else
    return HEIGHT_UNCHANGED;

  return special_case ? HEIGHT_UNCHANGED : HEIGHT_CHANGED;
}

/* Finds the item in tree ROOTP having the minimal value. */
static void *
find_min (avl_node * tree)
{
  if (!tree)
    return NULL;

  for (;;)
    {
      avl_node *left = tree->subtree[AVL_LEFT];
      if (!left)
	return tree;
      tree = left;
    }
}

/* Finds the item in tree ROOTP having the maximal value. */
static void *
find_max (avl_node * tree)
{
  if (!tree)
    return NULL;
  for (;;)
    {
      avl_node *right = tree->subtree[AVL_RIGHT];
      if (!right)
	return tree;
      tree = right;
    }
}

/* Finds item DATA in tree ROOTP, a subtree of cur_tree. */
static void *
find (void *data, avl_node * tree)
{
  int cmp;

  while (tree)
    {
      cmp = cur_tree->compare (data, tree->data, cur_tree->param);
      if (cmp == 0)
	return tree->data;
      tree = tree->subtree[cmp < 0 ? AVL_LEFT : AVL_RIGHT];
    }

  return NULL;
}

/* Inserts item *DATA in tree *ROOTP, a subtree of cur_tree.  Returns
   indication of whether the tree's height changed.  On exit DATA is
   set to NULL if insertion was successful, otherwise *DATA points to
   the duplicate. */
static int
insert (void **data, avl_node ** rootp)
{
  int increase;
  int cmp;

  if (*rootp == NULL)
    {
      *rootp = new_node (*data);
      *data = NULL;
      return HEIGHT_CHANGED;
    }

  cmp = cur_tree->compare (*data, (*rootp)->data, cur_tree->param);
  if (cmp < 0)
    {
      increase = -insert (data, &(*rootp)->subtree[AVL_LEFT]);
      if (*data)
	return HEIGHT_UNCHANGED;
    }
  else if (cmp > 0)
    {
      increase = insert (data, &(*rootp)->subtree[AVL_RIGHT]);
      if (*data)
	return HEIGHT_UNCHANGED;
    }
  else
    {
      *data = (*rootp)->data;
      return HEIGHT_UNCHANGED;
    }

  (*rootp)->bal += increase;

  /* Rebalance if needed.  Height of current tree increases only if
     its subtree height increases and the current tree needs no
     rotation. */
  return increase && (*rootp)->bal ? 1 - balance (rootp) : HEIGHT_UNCHANGED;
}

/* Replaces item *DATA in tree *ROOTP, a subtree of cur_tree.  That
   is, if item *DATA exists in ROOTP, it is replaced; if it does not,
   it is inserted.  On exit *DATA is set to the previous item, or NULL
   if no item was replaced.  Returns indication of whether the tree's
   height changed. */
static int
replace (void **data, avl_node ** rootp)
{
  int increase;
  int cmp;

  if (*rootp == NULL)
    {
      *rootp = new_node (*data);
      *data = NULL;
      return HEIGHT_CHANGED;
    }

  cmp = cur_tree->compare (*data, (*rootp)->data, cur_tree->param);

  if (cmp < 0)
    {
      increase = -replace (data, &(*rootp)->subtree[AVL_LEFT]);
      if (*data != NULL)
	return HEIGHT_UNCHANGED;
    }
  else if (cmp > 0)
    {
      increase = replace (data, &(*rootp)->subtree[AVL_RIGHT]);
      if (*data != NULL)
	return HEIGHT_UNCHANGED;
    }
  else
    {
      void *temp = (*rootp)->data;
      (*rootp)->data = *data;
      *data = temp;
      return HEIGHT_UNCHANGED;
    }

  (*rootp)->bal += increase;


  /* Rebalance if needed.  Height of current tree increases only if
     its subtree height increases and the current tree needs no
     rotation. */
  return increase && (*rootp)->bal ? 1 - balance (rootp) : HEIGHT_UNCHANGED;
}

/* Removes item *DATA from tree *ROOTP, a subtree of cur_tree.
   Returns indication of whether the trees height changed.  Points
   *DATA to the deleted item or NULL if none was found.

   Yup, the name is a C++ operator.  So what? */
static int
delete (void **data, avl_node ** rootp, avl_comparison_func compare)
{
  int decrease;
  int cmp;
  avl_node *old_root = *rootp;
  NODE nd_typ = node_type (*rootp);
  DIRECTION dir = (nd_typ == IS_LBRANCH) ? AVL_LEFT : AVL_RIGHT;

  if (*rootp == NULL)
    {
      *data = NULL;
      return HEIGHT_UNCHANGED;
    }

  cmp = ((int (*)()) compare) (*data, (*rootp)->data, cur_tree->param, nd_typ);

  if (cmp < 0)
    {
      decrease = -delete (data, &(*rootp)->subtree[AVL_LEFT], compare);
      if (*data == NULL)
	return HEIGHT_UNCHANGED;
    }
  else if (cmp > 0)
    {
      decrease = delete (data, &(*rootp)->subtree[AVL_RIGHT], compare);
      if (*data == NULL)
	return HEIGHT_UNCHANGED;
    }
  else
    {
      /* *ROOTP is the node to delete.  Cases:

         1. The node is a leaf.  Remove and return.

         2. The node is a branch (only one child).  Make *ROOTP, the
         pointer to this node, point to the child.

         3. The node has two children.  Swap data with the successor of
         *ROOTP (the smallest item in its right subtree) and delete
         that node.  `decrease' is set if the subtree height decreased
         due to this.  */

      *data = (*rootp)->data;
      switch (nd_typ)
	{
	case IS_LEAF:
	  if (cur_tree->owner == NULL)
	    free (*rootp);
	  *rootp = NULL;
	  return HEIGHT_CHANGED;
	case IS_RBRANCH:
	case IS_LBRANCH:
	  *rootp = (*rootp)->subtree[dir];
	  if (cur_tree->owner == NULL)
	    free (old_root);
	  return HEIGHT_CHANGED;
	case IS_TREE:
	  decrease = delete (&(*rootp)->data,
			     &(*rootp)->subtree[AVL_RIGHT],
			     (avl_comparison_func) min_node);
	  break;
	case IS_NULL:
	  ;
	}
    }

  (*rootp)->bal -= decrease;

  /* Rebalance if necessary: (1) a rotation changed the height of the
     subtree, or (2) the subtree height decreased and now matches the
     height of its other subtree so that the current tree now has a
     zero balance when previously it did not. */
  if (decrease && (*rootp)->bal)
    return balance (rootp);
  else if (decrease && !(*rootp)->bal)
    return HEIGHT_CHANGED;
  else
    return HEIGHT_UNCHANGED;
}

/* Tree walkers. */

/* Traverses level LEVEL of tree TREE calling WALK_FUNC in
   preorder. */
static void
walk_preorder (avl_node * tree, avl_walk_func walk_func, int level)
{
  walk_func (tree->data, cur_tree->param, level);

  if (tree->subtree[AVL_LEFT])
    walk_preorder (tree->subtree[AVL_LEFT], walk_func, level + 1);
  if (tree->subtree[AVL_RIGHT])
    walk_preorder (tree->subtree[AVL_RIGHT], walk_func, level + 1);
}

/* Traverses level LEVEL of tree TREE calling WALK_FUNC in
   in-order. */
static void
walk_inorder (avl_node * tree, avl_walk_func walk_func, int level)
{
  if (tree->subtree[AVL_LEFT])
    walk_inorder (tree->subtree[AVL_LEFT], walk_func, level + 1);

  walk_func (tree->data, cur_tree->param, level);

  if (tree->subtree[AVL_RIGHT])
    walk_inorder (tree->subtree[AVL_RIGHT], walk_func, level + 1);
}

/* Frees up space for all nodes in tree TREE (a subtree of cur_tree),
   calling FREE_FUNC in in-order (if non-NULL). */
static void
destroy (avl_node * tree, avl_free_func free_func)
{
  if (tree->subtree[AVL_LEFT])
    destroy (tree->subtree[AVL_LEFT], free_func);

  if (free_func)
    free_func (tree->data, cur_tree->param);

  if (tree->subtree[AVL_RIGHT])
    destroy (tree->subtree[AVL_RIGHT], free_func);

  if (cur_tree->owner == NULL)
    free (tree);
}

/* Public functions. */

/* Creates an AVL tree that allocates inside arena *OWNER, compares
   keys with COMPARE, with user data PARAM. */
avl_tree *
avl_create (arena ** owner, avl_comparison_func compare, void *param)
{
  avl_tree *tree;

  if (owner)
    tree = (avl_tree *) arena_alloc (owner, sizeof (avl_tree));
  else
    tree = (avl_tree *) xmalloc (sizeof (avl_tree));

  tree->owner = owner;
  tree->root = NULL;
  tree->compare = compare;
  tree->count = 0;
  tree->param = param;

  return tree;
}

/* Used by avl_sort(). */
static avl_tree *new_tree;

/* Adds each node in the old tree to the new tree as a helper function
   to avl_sort. */
static void
add_to_new_tree (void *data, void *param, int level)
{
  avl_insert (new_tree, data);
}

/* Creates an AVL tree with nodes identical to OLD_TREE except for
   different arena OWNER, comparison function COMPARE, and user data
   PARAM.  Returns the new tree.  The old tree is destroyed (but not
   its data).  The effect is that tree OLD_TREE is re-sorted using the
   function COMPARE. */
avl_tree *
avl_sort (avl_tree * old_tree, arena ** owner, avl_comparison_func compare,
	  void *param)
{
  new_tree = avl_create (owner, compare, param);
  avl_walk_inorder (old_tree, add_to_new_tree, NULL);
  avl_destroy (old_tree, NULL);
  return new_tree;
}

/* Destroys AVL tree TREE.  Frees up all associated space if not in an
   arena.  In any case calls FREE_FUNC for each node in tree TREE in
   in-order. */
void
avl_destroy (avl_tree * tree, avl_free_func free_func)
{
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  if (tree->root)
    destroy (tree->root, free_func);
  cur_tree = save_cur_tree;

  if (tree->owner == NULL)
    free (tree);
}

/* Traverses tree TREE, performing action WALK_FUNC on each item in
   the tree in preorder. */
void
avl_walk_preorder (avl_tree *tree, avl_walk_func walk_func, void *param)
{
  void *save_cur_tree = cur_tree;
  void *save_param = tree->param;

  if(tree->root == NULL)
    return;
  
  cur_tree = tree;
  tree->param = param;
  walk_preorder (tree->root, walk_func, 1);
  cur_tree = save_cur_tree;
  tree->param = save_param;
}

/* Traverses tree TREE, performing action WALK_FUNC on each item in
   the tree in in-order. */
void
avl_walk_inorder (avl_tree * tree, avl_walk_func walk_func, void *param)
{
  void *save_cur_tree = cur_tree;
  void *save_param = tree->param;

  if(tree->root == NULL)
    return;
  
  cur_tree = tree;
  tree->param = param;
  walk_inorder (tree->root, walk_func, 1);
  cur_tree = save_cur_tree;
  tree->param = save_param;
}

/* Traverses tree TREE in in-order.  If *TRAVP is NULL, then
   *initializes *TRAVP; otherwise uses *TRAVP as state.  Returns the
   *next node in the tree in in-order; after the last node, returns
   *NULL and TRAVP is reset to NULL (state is freed).  Uses malloc
   regardless of whether TREE has an associated arena.  */
void *
avl_traverse (avl_tree * tree, avl_traverser ** travp)
{
  avl_traverser *trav;
  avl_node *t;

  if (*travp == NULL)
    {
      trav = *travp = xmalloc (sizeof (avl_traverser));
      trav->nstack = 0;
      trav->p = tree->root;
    }
  else
    trav = *travp;

  /* Traversal algorithm borrowed from _The Art of Computer
     Programming_, Vol. 1, 2.3.1, Algorithm T.  */
  while (1)
    {
      while (trav->p)
	{
	  trav->stack[trav->nstack++] = trav->p;
	  trav->p = trav->p->subtree[AVL_LEFT];
	}
      if (!trav->nstack)
	{
	  free (trav);
	  *travp = NULL;
	  return NULL;
	}
      trav->p = trav->stack[--trav->nstack];
      t = trav->p;
      trav->p = trav->p->subtree[AVL_RIGHT];
      return t->data;
    }
}

/* Returns the number of nodes in tree TREE. */
int
avl_count (avl_tree * tree)
{
  assert (tree);
  return tree->count;
}

/* Inserts item DATA into tree TREE.  Returns NULL if the item was
   inserted, otherwise a pointer to the duplicate item. */
void *
avl_insert (avl_tree * tree, void *data)
{
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  insert (&data, &tree->root);
  cur_tree = save_cur_tree;

  if (data == NULL)
    tree->count++;

  return data;
}

/* If item DATA does not exist in TREE, inserts it.  If the item does
   exist, it is replaced by DATA.  Returns the item replaced or NULL
   if inserted.  The calling function should free the returned item if
   appropriate. */
void *
avl_replace (avl_tree * tree, void *data)
{
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  replace (&data, &tree->root);
  cur_tree = save_cur_tree;

  if (data == NULL)
    tree->count++;

  return data;
}

/* Deletes item DATA from tree TREE.  If the item exists in TREE,
   returns that item; otherwise returns NULL.  The calling function
   should free the returned item if appropriate. */
void *
avl_delete (avl_tree * tree, void *data)
{
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  delete (&data, &tree->root, tree->compare);
  cur_tree = save_cur_tree;

  if (data)
    tree->count--;

  return data;
}

/* Finds item DATA in tree TREE and returns it or NULL if not
   found. */
void *
avl_find (avl_tree * tree, void *data)
{
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  data = find (data, tree->root);
  cur_tree = save_cur_tree;

  return data;
}

/* Deletes the minimal item from tree TREE.  Returns the deleted item
   or NULL if the tree was empty.  The calling function should free
   the returned item if appropriate. */
void *
avl_delete_min (avl_tree * tree)
{
  void *data;
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  delete (&data, &tree->root, (avl_comparison_func) min_node);
  cur_tree = save_cur_tree;

  if (data)
    tree->count--;

  return data;
}

/* Returns the minimal item in tree TREE, or NULL if the tree is
   empty. */
void *
avl_find_min (avl_tree * tree)
{
  return tree->root == NULL ? NULL : find_min (tree->root);
}

/* Deletes the maximal item from tree TREE.  Returns the deleted item
   or NULL if the tree was empty.  The calling function should free
   the returned item if appropriate. */
void *
avl_delete_max (avl_tree * tree)
{
  void *data;
  void *save_cur_tree = cur_tree;
  cur_tree = tree;
  delete (&data, &tree->root, (avl_comparison_func) max_node);
  cur_tree = save_cur_tree;

  if (data)
    tree->count--;

  return data;
}

/* Returns the maximal item in tree TREE, or NULL if the tree is
   empty. */
void *
avl_find_max (avl_tree * tree)
{
  return tree->root == NULL ? NULL : find_max (tree->root);
}

#if GLOBAL_DEBUGGING
/* Checks that avl_delete() returns a non-NULL value. */
void *
force_avl_delete (avl_tree * tree, void *p)
{
  void *r = avl_delete (tree, p);
  assert (r);
  return r;
}
#endif

static void
shallow_copy (void *data, void *param, int level)
{
  avl_insert ((avl_tree *) param, data);
}

/* Copies avl_tree TREE and returns a pointer to the copy.  The
   pointers to the items in the tree are copied, not the items
   themselves.  The copied tree is not owned by an arena. */
avl_tree *
avl_shallow_copy (avl_tree *tree)
{
  avl_tree *copy = xmalloc (sizeof (avl_tree));
  copy->owner = NULL;
  copy->root = NULL;
  copy->compare = tree->compare;
  copy->count = 0;
  copy->param = tree->param;
  
  avl_walk_preorder (tree, shallow_copy, copy);

  return copy;
}
