/* PSPP - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#if !tab_h
#define tab_h 1

#include <limits.h>

/* Cell options. */
enum
  {
    TAB_NONE = 0,

    /* Must match output.h: OUTP_T_JUST_*. */
    TAB_ALIGN_MASK = 03,	/* Alignment mask. */
    TAB_RIGHT = 00,		/* Right justify. */
    TAB_LEFT = 01,		/* Left justify. */
    TAB_CENTER = 02,		/* Center. */
    TAB_JUSTIFY = 03,		/* Fully justify. */

    /* Must match output.h:OUTP_T_FANCY. */
    TAB_RICH = 04,		/* Rich text. */

    TAB_JOIN = 010,		/* Joined cell. */
  };

/* Line styles.  These must match output.h:OUTP_L_*. */
enum
  {
    TAL_0 = 0,			/* No line. */
    TAL_1THIN = 0,		/* Single line, if it's space-efficient. */
    TAL_1 = 1,			/* Single line. */
    TAL_2 = 2,			/* Double line. */
    TAL_3 = 3,			/* Special line of driver-defined style. */
    TAL_COUNT,			/* Number of line styles. */

    TAL_SPACING = 0200,		/* Don't draw the line, just reserve space. */
  };

/* Column width and row height operators. */
enum
  {
    t_end = INT_MIN,

    /* Font measurements. */
    t_psw,			/* Proportional font string width. */
    t_ptw,			/* Proportional title font string width. */
    t_fsw,			/* Fixed font string width. */
    t_pcw,			/* Proportional font em width. */
    t_fcw,			/* Fixed font character width. */
    t_hgt,			/* Font height. */

    /* Table attributes. */
    t_nr,			/* Number of rows. */
    t_nc,			/* Number of columns. */

    /* Other measurements. */
    t_col,			/* Column width. */
    t_row,			/* Row height. */
    t_nah,			/* Natural height, expensive! */
    t_naw,			/* Natural width, expensive! */

    /* Basic operations. */
    t_con,			/* Constant. */
    t_add,			/* Addition. */
    t_sub,			/* Subtraction. */
    t_mul,			/* Multiplication. */
    t_div,			/* Division. */
    t_neg,			/* Negation. */
    t_max,			/* Maximum. */
    t_min,			/* Minimum. */
    t_xch,			/* Exchange. */
    t_dup,			/* Duplicate. */

    /* Tests and loops. */
    t_lbl,			/* Label. */
    t_jnz,			/* Jump if not zero. */
      
    /* Optional attributes. */
    t_str,			/* Maximum stretch amount. */
    t_shr,			/* Maximum shrink amount. */

    /* Saving results. */
    t_sac,			/* Save column. */
    t_sar,			/* Save row. */
    t_scr,			/* Save column range. */
    t_srr,			/* Save row range. */

    t_sentinel,
  };

/* Joined cell. */
typedef struct tab_joined_cell
  {
    int x1, y1;
    int x2, y2;
    int hit;
    a_string contents;
  }
tab_joined_cell;

/* A table. */
typedef struct tab_table
  {
    struct arena *container;
    
    /* Columns. */
    int col_style;		/* One of SOM_COL_*. */
    int col_group;		/* Number of rows per group. */
    
    /* Table info. */
    a_string title;		/* Table title. */
    unsigned flags;		/* SOMF_*. */
    int nc, nr;			/* Number of columns, rows. */
    int l, r, t, b;		/* Number of header rows on each side. */
    int wl, wr, ht, hb;		/* Width/height of header rows/columns. */
    a_string *cc;		/* Cell contents; a_string *[nr][nc]. */
    unsigned char *ct;		/* Cell types; unsigned char[nr][nc]. */
    unsigned char *rh;		/* Horiz rules; unsigned char[nr+1][nc]. */
    int *trh, *hrh;		/* Types of, heights of horiz rules; [nr+1]. */
    unsigned char *rv;		/* Vert rules; unsigned char[nr][nc+1]. */
    int *trv, *wrv;		/* Types of, widths of vert rules; [nc+1]. */
    unsigned char *dim;		/* Expressions for column width, row height. */
    int max_stack_height;	/* Max stack height for evaluating `dim'. */
    som_dimension *w, *h;	/* Column widths, row heights. */
  }
tab_table;

extern som_table_class tab_table_class;
extern int tab_hit;

/* Tables. */
tab_table *tab_create (int nc, int nr, int resizable);
void tab_destroy (tab_table *);
void tab_height (tab_table *, int nr);
void tab_realloc (tab_table *, int nc, int nr);
void tab_headers (tab_table *, int l, int r, int t, int b);
void tab_columns (tab_table *, int style, int group);
void tab_title (tab_table *, int format, const char *, ...);
void tab_dim (tab_table *, ...);
void tab_nat_dim (tab_table *);
void tab_flags (tab_table *, unsigned);
void tab_submit (tab_table *);

/* Rules. */
void tab_hline (tab_table *, int style, int x1, int x2, int y);
void tab_vline (tab_table *, int style, int x, int y1, int y2);
void tab_box (tab_table *, int f_h, int f_v, int i_h, int i_v,
	      int x1, int y1, int x2, int y2);

/* Text options, passed in the `opt' argument. */
enum
  {
    TAT_NONE = 0,		/* No options. */
    TAT_PRINTF = 0x0100,	/* Format the text string with sprintf. */
    TAT_TITLE = 0x0204,		/* Title attributes (implies TAB_RICH). */
    TAT_FIX = 0x0400,		/* Fixed font (exclusive of TAB_RICH). */
    TAT_NOWRAP = 0x0800,        /* No text wrap (tab_output_text() only). */
  };

/* Cells. */
void tab_null (tab_table *, int c, int r);
void tab_value (tab_table *, int c, int r, unsigned char opt,
		    const value *, const fmt_spec *);
void tab_float (tab_table *, int c, int r, unsigned char opt,
		    double v, int w, int d);
void tab_text (tab_table *, int c, int r, unsigned opt, const char *, ...)
     __attribute__ ((format (printf, 5, 6)));
void tab_joint_text (tab_table *, int x1, int y1, int x2, int y2,
		     unsigned opt, const char *, ...)
     __attribute__ ((format (printf, 7, 8)));

/* Simple output. */
void tab_output_text (int options, const char *string, ...)
     __attribute__ ((format (printf, 2, 3)));

#endif /* tab_h */

